/*
 * Simple trace program for the Alpha.
 * Originally written by James Bonfield. (jkb@mrc-lmb.cam.ac.uk)
 * 
 * Currently being maintained by Anthony Baxter <anthony@aaii.oz.au>.
 *
 * To use please make sure you have the /proc file system mounted by either
 * including the following line in /etc/fstab:
 * "/proc           /proc           procfs  rw              0 0"
 * or by mounting with the command:
 * "mount -t procfs /proc /proc"
 *
 * Version 1.0beta
 *
 * Last update 11/10/93
 * $Id: trace.c,v 1.20 1993/11/22 12:28:20 anthony Exp $
 */

#include "trace.h"
#include "signame.h"
#include "sysname.h"
#include "digital_alpha_reg.h"

#define USE_PR_HEX

int max_string = 40;

#define proc_cntl(a,b) \
    if (-1 == ioctl(p_fd, a, b)) \
	die(#a);

typedef struct
{
	int print_pid;
	int follow_fork;
	int do_out_files;
	int pid_supplied;
    char output_template[MAXPATHLEN];
} tstOpt;

static pid_t pid = 0;
static int gotalarm = 0;
int p_fd = 0;
FILE *outfp = stderr;
static void catchalarm();

static struct prrun pr = 
	{
		PRSTRACE,
		0, 0, 0, 0, 0, 0
	};

/**
 * exit with a message
 */
void die(char *msg) 
{
    if(msg)  perror(msg);
    if(pid)  kill(pid, SIGKILL);
    if(p_fd) close(p_fd);
    exit(1);
}

/**
 * verify the syscall table against actual values
 * for the OS on which the sources were compiled.
 * Serves as a check against modified system call
 * table.
 */
void check_syscall_table(void)
{
	int	i;

	for (i = 0; i < Nsyscalls; ++i) 
	{
		if(sysent[i].syscallnum == i) continue;
		if(sysent[i].syscallnum == SYS_unknown) continue;
		if(sysent[i].syscallnum == SYS_obsolete) continue;
		fprintf(stderr,
		    "Warning: sysent[%03.3d] has syscall %03.3d %s\n",
		    i, sysent[i].syscallnum, sysent[i].returns);
	}
	/* check numargs vs name(AAA)r ? */
}

/**
 * prints from the buffer max_len bytes.
 * if not fixed and the complete string was not
 * printed, prints a continuation symbol "..."
 * at the end
 */
void pr_string(char *strp, int max_len, int fixed) 
{
    char c;

    fputc('"', outfp);
    while ((c = *strp++) && max_len--) 
	{
		switch(c) 
		{
			case '"':  fprintf(outfp,"\\\""); break;
			case '\n': fprintf(outfp,"\\n"); break;
			case '\t': fprintf(outfp,"\\t"); break;
			case '\r': fprintf(outfp,"\\r"); break;
			case '\\': fprintf(outfp,"\\"); break;
			default:
				if (isascii(c) && isprint(c)) 
				{
					fputc(c,outfp);
				} 
				else 
				{
					fprintf(outfp,"\\%03o", c);
				}
		}
    }
    fputc('"', outfp);
    if (!fixed && (max_len == -1) && c)
	{
		fprintf(outfp, "...");
	}
}

/**
 * prints from the buffer max_len bytes in hex
 * format.
 * if not fixed prints a continuation symbol "..."
 * at the end
 */
void pr_hex(char *strp, int max_len, int fixed) 
{
    char c;
    char s = '<';

    while(max_len--) 
	{
        c = *strp++;
        fprintf(outfp, "%c%02.2x", s, c & 255);
		s = ' ';
    }
    fputc('>', outfp);
    if (!fixed && (max_len == -1))
	{
		fprintf(outfp, "...");
	}
}

/**
 * reads a buffer from the file and 
 * returns back in a newly allocated chunk
 * of memory. The returned buffer needs to
 * be free'd by the caller.
 */
char *rd_buf(long val, size_t len) 
{
    char *str = (char *)malloc(len+1);

    if (!str) die("malloc");
    
    if (-1 == lseek(p_fd, val, SEEK_SET)) die("lseek p_fd");
    if (-1 == read(p_fd, str, len)) die("read p_fd");

    return str;
}

void p_arg(char cType, long val, long length) 
{
    switch(cType) 
	{
		case 'd':
			fprintf(outfp,"%ld", val);
			break;
		case 'u':
			fprintf(outfp,"%lu", val);		/* was %uld ?!?! */
			break;
		case 'p':
			fprintf(outfp,"0x%p", val);
			break;
		case 'o':
			fprintf(outfp,"0%o", val);
			break;
		case 's': 
			{
			char *str = rd_buf(val, max_string+1);

			pr_string(str, max_string, 0);
			free(str);
			break;
			}
		case 'v':
			fprintf(outfp,"(void)");
			break;
		case '?': 
			{
			char *str = rd_buf(val, max_string+1);
			int i, t = 0;
			
			for (i = 0; str[i] && i < max_string; i++) 
			{
				if (isascii(str[i]) && isprint(str[i]))
				{
					t++;
				}
			}

			if ((i-1) && (float)t/(i-1) > .8)
				pr_string(str, max_string, 0);
			else

#ifdef USE_PR_HEX
				pr_hex(str, length < max_string ? length : max_string, 0);
#else
				fprintf(outfp,"0x%p", val);
#endif
			free(str);
			break;
			}
		case 'b': 
			{
			int i, t = 0, tlen = length > max_string ? max_string : length;
			char *str = rd_buf(val, tlen+1);
			
			for (i = 0; i < tlen; i++) 
			{
				if (isascii(str[i]) && (isprint(str[i]) || isspace(str[i])))
				{
					t++;
				}
			}

			if (tlen == 0 || (tlen && (float)t/tlen > .8))
				pr_string(str, tlen, !(length > max_string));
			else
#ifdef USE_PR_HEX
				pr_hex(str, length < max_string ? length : max_string, 0);
#else
				fprintf(outfp,"0x%p", val);
#endif

			free(str);
			break;
			}
		default:
			fprintf(outfp,"?0x%p?", val);
			break;
    }
}

/* highest possible start address for a user process */
#define MAX_U_START 0x0000020000000000

/* argv: child process argument vector */
void do_exec(char *argv[]) 
{
    char buf[1024];
    int pfd, flag;
    sigset_t sigtrace;
    struct prrun prn;
    struct prpsinfo ps;
    struct sigvec vec, ovec;
    long omask;

    /* open yourself so tracing flags can be set for the child */
    sprintf(buf, "/proc/%d", getpid());
    if( ((pfd = open(buf, O_RDWR))) == -1) die("parent open");

    /* Set trace on any exec via PIOCSSPCACT, 
	 * Note: this automatically sets trace on signal SIGTRAP.
     */
    flag = PRFS_STOPEXEC;
    if(ioctl(pfd, PIOCSSPCACT, &flag) < 0) die("parent PIOCSSPCACT");

    /* Set inherit on fork flag so the child will get all our tracing flags.
     */
    if(ioctl(pfd, PIOCSFORK, 0) < 0) die("parent PIOCSFORK");

    /* Set up signal handlers...
     */
    vec.sv_handler = catchalarm; 
    vec.sv_mask = 0;
    vec.sv_onstack = 0;
    (void) sigvec(SIGALRM, &vec, &ovec);
    omask = sigblock(sigmask(SIGALRM));
    gotalarm = 0;

    /* fork a child process; child waits for a signal from parent, child
       execs path; child should stop on exec */
    pid = fork();
    switch(pid) 
	{
		case -1:			/* failure */
			die("fork");
		case 0:				/* child */
			while (!gotalarm)
			{
				sigpause(omask &~ sigmask(SIGALRM));
			}
			/* reset our signal handlers */
			(void)sigvec(SIGALRM, &ovec, (struct sigvec *)0);
			(void)sigsetmask(omask);
			execvp(argv[0], argv);
			/* you only reach here if the execvp fails */
			die("child execvp");
    }

    /* this is all parent code from here down */

    (void) sigvec(SIGALRM, &ovec, (struct sigvec *)0);
    (void) sigsetmask(omask);

    /* open the child task */
    sprintf(buf, "/proc/%d", pid);
    fprintf(outfp, "Tracing process %s\n", buf);
    if( ((p_fd = open(buf, O_RDWR))) == -1)
	{
        die(buf);
	}

    /* wake the child up */
    flag=SIGALRM;
    proc_cntl(PIOCKILL,&flag);   

    /* wait for child process to stop at the first exec */
    proc_cntl(PIOCWSTOP, 0);

    /* Clear all tracing flags in the parent (pfork) and close its pfd */

    /* clear trace on signal */
    premptyset(&sigtrace);
    if (ioctl(pfd, PIOCSTRACE, &sigtrace) < 0) die("parent PIOCSTRACE");

    /* clear trace on any exec via PIOCSSPCACT */
    flag = 0;
    if (ioctl(pfd, PIOCSSPCACT, &flag) < 0) die("parent PIOCSSPCACT");
    if (close(pfd) < 0) die("parent close");

    proc_cntl(PIOCPSINFO, &ps);

    /* If the address is not in the range of a "normal" user process, we
     * must be in the loader, so give the process a run command to get it
     * to the actual user code.
     */
    if ((long)(ps.pr_addr) >= MAX_U_START) 
	{
		/* The child is stopped in the user loader, run it, and wait
		 * for it to stop. This stop will be in the actual user program.
		 */
		prn.pr_flags = PRCSIG;
		proc_cntl(PIOCRUN, &prn);
		proc_cntl(PIOCWSTOP, 0);
	}
    /* now stopped at the real child process exec */

    /* Clear the inherit on fork flag in the child */
    proc_cntl(PIOCRFORK, 0);

    prn.pr_flags = PRCSIG;
    proc_cntl(PIOCRUN, &prn);
}

static void catchalarm() 
{
    gotalarm=1;
    return;
}

/**
 * get command line options and put them into the supplied struct
 */
void get_options(tstOpt *pstOpt, int argc, char **argv)
{
	int c, usage = 0;
    /* parse arguments */
    extern char *optarg;

	memset(pstOpt, 0, sizeof(*pstOpt));

    while (-1 != (c = getopt(argc, argv, "fPo:p:s:O:"))) 
	{
		switch (c) 
		{
			case 'o':
				if (NULL == (outfp = fopen(optarg, "w"))) die(optarg);
				break;
			case 'p':
				pid = atoi(optarg);
				pstOpt->pid_supplied = 1;
				break;
			case 's':
				max_string = atoi(optarg);
				break;
			case 'P':
				pstOpt->print_pid = 1; 
				break;
			case 'f':
				pstOpt->follow_fork = 1; 
				break;
			case 'O':
				strcpy(pstOpt->output_template,optarg);
				pstOpt->do_out_files=1;
				break;
			case '?':
				usage++;
		}
    }

    if (usage || (pstOpt->pid_supplied == 0 && argc == 0)) 
	{
		fprintf(stderr,
		  "Trace for /proc - $Revision: 1.30 $ \n"
           "Usage: trace [-f] [-P] [-o outfile] [-O outfile_template] [-s string_length] command [arguments]\n"
           "       trace [-f] [-P] [-o outfile] [-O outfile_template] [-s string_length] -p pid\n");
		exit(1);
    }
}

void do_trace(tstOpt *pstOpt, prstatus_t *pprstat, sysset_t *pss)
{
	gregset_t regs;
	int wstat;
	int c_fd=0;
	int in_fork=0, in_execve=0, am_child=0;
	pid_t child_pid, tracerpid;
	char out_header[80];
	int flag;

	for(;;) 
	{
		proc_cntl(PIOCWSTOP, pprstat);
		*out_header='\0';
		if(pstOpt->print_pid) sprintf(out_header,"[%d]: ",pprstat->pr_pid);

		if (pprstat->pr_why == PR_SYSENTRY) 
		{
			proc_cntl(PIOCGREG, &regs);
			if (pprstat->pr_what == SYS_execve) 
			{
				in_execve=2;
			}
			if((pprstat->pr_what==SYS_fork) && (pstOpt->follow_fork)) 
			{
				in_fork=1;
				proc_cntl(PIOCSFORK, 0);
				proc_cntl(PIOCRRLC, 0); /* to stop child running post-fork */
			}
			fputs(out_header, outfp);

			/*
			 * new syscalls added by new OS releases used to cause us to
			 * `fall off the end of our table', now we catch these
			 */
			if (regs.regs[0] >= Nsyscalls) 
			{
				fprintf(outfp, "syscall#%u past end of table!\n", regs.regs[0]);
			} 
			else 
			{
				flag=(*sysent[regs.regs[0]].printargsfunc)(
					regs.regs[0],    
					&regs.regs[EF_A3],
					sysent[regs.regs[0]].numargs,
					sysent[regs.regs[0]].returns);
			}
		} 
		else if (pprstat->pr_why == PR_SYSEXIT) 
		{
			proc_cntl(PIOCGREG, &regs);

			/* The next bit is hairy. The system call has failed, and register 0   */
			/* contains errno rather than the return value, _if_ EF_T8 is 1        */
			/* Why EF_T8? Because thats how it works, as verified by disassembling */
			/* syscall.o :-)                                                       */
			if (pprstat->pr_what >= Nsyscalls) 
			{
				/* already bitched on entry, just ignore */
			} 
			else 
			{
				flag=(*sysent[pprstat->pr_what].printretfunc)(
					pprstat->pr_what,
					(regs.regs[EF_T8])?(long)(-1):(int)(regs.regs[0]),/*return*/
					(regs.regs[EF_T8])?(regs.regs[0]):(0));	      /*errno*/
			}
			/*
			 * The bit above was moved here and the fflush was added just below here
			 * to prevent the stdio library from outputing the current contents of its
			 * buffer twice around this fork (once in the parent and once in the child).
			 */
			fflush(outfp);

			if(in_fork) 
			{
				in_fork=0;
				proc_cntl(PIOCSRLC,0); /* reset run on last close in parent */
				if(!regs.regs[EF_T8]) 
				{ /* did fork() succeed? */
					pstOpt->print_pid=1;
					child_pid=regs.regs[0];
					proc_cntl(PIOCSRLC,0);
					tracerpid=fork();
					switch(tracerpid) 
					{
						case -1: die("trace couldnt fork()"); break;
						case  0: am_child=1; break;
						default: am_child=0; break;
					}
					if(am_child) 
					{
						char buf[1024];

						sprintf(buf, "/proc/%05d", child_pid);
						if (-1 == (c_fd = open(buf, O_RDWR | O_EXCL, 0))) die(buf);
						close(p_fd);
						p_fd=c_fd;
						pid=child_pid;
						if(pstOpt->do_out_files) 
						{
							char outbuf[MAXPATHLEN];
							sprintf(outbuf,"%s%05d",pstOpt->output_template,child_pid);
							if(NULL == (outfp=fopen(outbuf,"w")) ) 
							die(outbuf);
						}
						fprintf(outfp,"Tracing process %s\n",buf);
						prfillset(pss);
						prfillset(&pr.pr_trace);
						proc_cntl(PIOCSENTRY, pss);
						proc_cntl(PIOCSEXIT, pss);
						proc_cntl(PIOCGSPCACT,&flag);
						flag |= PRFS_STOPTERM;
						flag ^= PRFS_STOPEXEC; /* turn off stop on exec */
						proc_cntl(PIOCSSPCACT, &flag);

						/* arrange for all signals to be traced */
						proc_cntl(PIOCSTRACE, &pr.pr_trace);
						proc_cntl(PIOCSRLC, 0);
						continue;
					}
				}  /* fork() failed */
			}   /* end of fork() handling code */
		} 
		else if (pprstat->pr_why == PR_SIGNALLED) 
		{
			/* normal case */
			fprintf(outfp, "%sSIGNAL [%d %s]\n", out_header, 
				pprstat->pr_what, sig_name[pprstat->pr_what]);

			if (pprstat->pr_what == SIGTRAP) 
			{
				/*
				 * If we're in execve then the SIGTRAP has been caused due
				 * to PIOCSSPCACT/PRFS_STOPEXEC. We should not let this
				 * signal get through to the child. We should also make sure
				 * we repeat this operation until we've done this in the
				 * actual user (rather than loader) program.
				 */
				if (in_execve) 
				{
					struct prrun npr = pr;
					struct prpsinfo ps;
					
					proc_cntl(PIOCPSINFO, &ps);
					if ((long)(ps.pr_addr) >= MAX_U_START) 
					{
						/* Ideally should just be in_execve=0, but for some
						 * reason, we are above MAX_U_START even when in the
						 * loader
						 */
						in_execve--;
					}
					npr.pr_flags = PRCSIG;
					proc_cntl(PIOCRUN, &npr);
				} 
				else 
				{
					/* hack to prevent unfinite loops when SIGTRAP comes
					   in; needed for child process only; peer process OK */
					for (flag = 1; flag <= NSIG; flag++) 
					{
						if (prismember(&(pprstat->pr_sigpend), flag-1)) 
						{
#ifdef _SYS_SIGINFO_H /* future OSF/1 change reqd. */
							siginfo_t sig;
#else
							struct siginfo sig;
#endif
							/* signal 'i' is pending */
							fprintf(outfp, "%sSIGNAL [%d %s]\n", out_header, flag, sig_name[flag]);

							/* set this signal up as the current signal so it
							   is sent to the process being traced */
#ifdef _SYS_SIGINFO_H /* future OSF/1 change reqd. */
							sig.si_signo = flag;
#else
							sig.fill = flag;
#endif
							proc_cntl(PIOCSSIG, &sig);
				
							/* remove signal from the pending list */
							proc_cntl(PIOCUNKILL, &flag);
				
							/* only deal with one signal at each stop */
							break;
						}
					}
				}
			}
		} 
		else if (pprstat->pr_why == PR_FORKSTOP) 
		{
			/* child just completing a fork */
			/* the PIOCRUN below is all we need */
		} 
		else if (pprstat->pr_why == (short)PR_DEAD) 
		{
			fprintf(outfp, "%sEXIT [%s=%d]\n", out_header, 
				WIFEXITED(pprstat->pr_what)
					? "status"
					: "signal",
				WIFEXITED(pprstat->pr_what)
					? WEXITSTATUS(pprstat->pr_what)
					: WTERMSIG(pprstat->pr_what));
			break;
		} 
		else 
		{
			fprintf(outfp, "UNRECOGNISED PIOCWSTOP 0x%hx\n", pprstat->pr_why);
		}

		/*
		 * cleanup any of our zombies
		 */
		do {} while (waitpid(-1, &wstat, WNOHANG) > 0);

		proc_cntl(PIOCRUN, &pr);
	}
}

int main(int argc, char *argv[]) 
{
	tstOpt stOpt;
    prstatus_t prstat;
    sysset_t ss;
    extern int optind;

	get_options(&stOpt, argc, argv);
    argc -= optind;
    argv += optind;

    check_syscall_table();

    if (stOpt.pid_supplied) 
	{
		char buf[1024];

		/* trace already running process */
		sprintf(buf, "/proc/%05d", pid);
		fprintf(outfp, "Tracing process %s\n", buf);

		if (-1 == (p_fd = open(buf, O_RDWR | O_EXCL, 0))) die(buf);
    } 
	else 
	{
		/* start child process */
        printf("%s\n", argv[0]);
		do_exec(argv);
    }

    if(stOpt.do_out_files) 
	{
		char outbuf[MAXPATHLEN];
		proc_cntl(PIOCSTATUS, &prstat);
		sprintf(outbuf, "%s%05d", stOpt.output_template, prstat.pr_pid);
		if(NULL == (outfp=fopen(outbuf,"w")) ) die(outbuf);
    }

    /* arrange for the run on last close flag to be set in the traced
       process; this ensures it resumes processing normally when the
       trace program exits */
    proc_cntl(PIOCSRLC, &ss);

    /* arrange for stop on entry and exit of all system calls */
    prfillset(&ss);
    prfillset(&pr.pr_trace);
    proc_cntl(PIOCSENTRY, &ss);
    proc_cntl(PIOCSEXIT, &ss);
	{
		/* set flag to stop on termination */
		int flag;
		proc_cntl(PIOCGSPCACT,&flag);
		flag |= PRFS_STOPTERM;
		proc_cntl(PIOCSSPCACT, &flag);
	}

    /* arrange for all signals to be traced */
    proc_cntl(PIOCSTRACE, &pr.pr_trace);

    if(stOpt.follow_fork)
		proc_cntl(PIOCSFORK, 0);

    /* child process will always be stopped at this point; a peer process
       may or may not be stopped; however, PIOCWSTOP doesn't seem to
       mind being called when the traced process is already stopped */

    /* now do the tracing */
	do_trace(&stOpt, &prstat, &ss);

    die(0);
}

